/*********************************************
 * Test of a 20x4 LCD display directly attached to processor
 *
 * This program writes the hex digits 0-F on the first line,
 * and Hello World! on the second line of the LCD.
 *
 * PORTA - LCD Data pins DB0-DB7
 * PC5   - Enable
 * PC4   - R/W  L:Data write from MPU to LCM  H:Data read from LCM to MPU
 * PC3   - R/S  L:Instruction code input      H:Data Input
 *
 * Status LED on pin PC0
 *
 * *** NOTE
 * *** NOTE The JTAGEN fuse must be turned off or ports PC4 and PC5 will
 * *** NOTE have pullups enabled.
 * *** NOTE 
 *
 * Chip type           : ATMEGA162
 * Clock frequency     : 4.000000 MHz
 *********************************************/

/*************************************************************************
 *			Macros
 *************************************************************************/
/* Set CPU frequency for libc routines and macros */
#define F_CPU 4000000UL  // 4 MHz

#define SET_ENABLE   PORTC |= _BV(PC5);
#define CLEAR_ENABLE PORTC &= ~(_BV(PC5));

//#define TOGGLE_ENABLE SET_ENABLE ; delay_us(1); CLEAR_ENABLE;
#define TOGGLE_ENABLE SET_ENABLE ; delay_us(20); CLEAR_ENABLE;

#define SET_RW       PORTC |= _BV(PC4);
#define CLEAR_RW     PORTC &= ~(_BV(PC4));

#define SET_RS       PORTC |= _BV(PC3);
#define CLEAR_RS     PORTC &= ~(_BV(PC3));

#define STATUS_LED_ON 		PORTC |= _BV(PC0);
#define STATUS_LED_OFF 		PORTC &= ~(_BV(PC0));

/*************************************************************************
 *			Includes
 *************************************************************************/
#include <io.h>
#include <avr/delay.h>


/*************************************************************************
 *			Function Prototypes
 *************************************************************************/

static void delay_ms (uint16_t ms);
static void delay_us (int32_t us);
static int check_bf( void );

/*************************************************************************
 *			Functions
 *************************************************************************/


/*
  delay_ms
  --------
  Delay up to 2^16 milliseconds using the avr libc delay functions
*/
static void delay_ms (uint16_t ms)
{
	for (;ms>0;ms--)
		_delay_ms(1);
}


/*
  delay_us
  --------
  Delay up to 2^16 microseconds using the avr libc delay functions.
  Accurate to about 10 us.
*/
static void delay_us (int32_t us)
{
	for (;us>0;us--)
		_delay_us(1);
}

/*
  write_lcd_clear 
  ---------------
  Clear the display 
*/
void lcd_clear (void)
{

	while (check_bf() == 1)
		;

	CLEAR_ENABLE;
	CLEAR_RS;
	CLEAR_RW;
	delay_us(1);
	
	// Setup the data to send
	PORTA=0x01;

	// Tell the LCD to do it.
	TOGGLE_ENABLE;

	// This command takes almost 2ms to complete	
}

/*
  write_lcd_home
  ---------------
  Move cursor to the home position
*/
void lcd_home (void)
{
	while (check_bf() == 1)
		;
	
	CLEAR_ENABLE;
	CLEAR_RS;
	CLEAR_RW;
	delay_us(1);

	// Setup the data to send
	PORTA=0x02;

	// Tell the LCD to do it.	
	TOGGLE_ENABLE;

	// This command takes almost 2ms to complete

	
}

/*
  Write a character at the next position.
*/
void lcd_write_char(int8_t value)
{
	while (check_bf() == 1)
		;

	
	// Set RS high
	// Set RW low to write
	CLEAR_ENABLE;
	SET_RS;
	CLEAR_RW;
	delay_us(1);

	// Setup the data to send
	PORTA=value;

	// Tell the LCD to do it.	
	TOGGLE_ENABLE;
	
	// This operation takes 46us to complete
}



/*
   Sets the position of the cursor for writing the next character.
   The datasheet calls this the DD RAM address.
 */
void lcd_set_char_pos(int8_t row, int8_t col)
{
	int8_t row_data_address = 0x80;
	
	// Set RS high
	// Set RW low to write
	CLEAR_ENABLE;
	CLEAR_RS;
	CLEAR_RW;
	delay_us(1);

	// Setup the data to send
	switch (row)
	{
	case 0:
		row_data_address = 0x80;
		break;
	case 1:
		row_data_address = 0xC0;
		break;		
	case 2:
		row_data_address = 0x94;
		break;		
	case 3:
		row_data_address = 0xD4;
		break;
	}
	
	PORTA= row_data_address + col;

	// Tell the LCD to do it.	
	TOGGLE_ENABLE;
	
	// This operation takes 42us to complete

}

void lcd_write_string (char * string)
{
	char *p;
	for (p=string;*p;p++)
		lcd_write_char (*p);
}

/* Write a 5 digit string to the output */
void lcd_write_uint16 (uint16_t val, int8_t maxlen)
{
	char buf[6];
	char *p;
	memset(buf,0,sizeof(buf));

	for (p=&buf[4];p>=buf;p--)
	{
		*p=(val%10)+'0';
		val = val /10;
	}
	if (maxlen>0 && maxlen <5)
	{
		lcd_write_string(buf+(5-maxlen));
	}
	else
	{
		lcd_write_string(buf);
	}
}


/* function_mode_set
   --------------
   Tells LCD controller about the display
   eight_bit_mode: 0 == 4 it mode 1 == 8 bit mode
   display_lines:  0 == 1 line    1 == 2 lines
   font:           0 == 5x7       1 == 5x10
*/
static void function_mode_set
(
	int8_t eight_bit_mode,
	int8_t  display_lines,
	int8_t font
)
{
	uint8_t data_to_write = 0x20;
	
	CLEAR_ENABLE;
	CLEAR_RS; CLEAR_RW;
	delay_us(1);
	
	if (eight_bit_mode)
		data_to_write |= 0x10;
	if (display_lines)
		data_to_write |= 0x08;
	if (font)
		data_to_write |= 0x04;	
	
	PORTA=data_to_write;
	TOGGLE_ENABLE;
	
	// This operation takes 42us to complete
	delay_us(42);
}

/*
  lcd_set_display
  ---------------
  Write the command to turn the display on or off
  
  on_val:       1 == turn LCD on     0 == turn LCD off
  cursor_on:    1 == turn cursor on  0 == turn cursor off
  cursor_blink: 1 == blink on        0 == blink off
 */
static void lcd_set_display
(
	int8_t on_val,
	int8_t cursor_on,
	int8_t cursor_blink
)
{
	uint8_t data_to_write = 0x08;
/* ------------------------------------------------------------------------- */
	
	while (check_bf() == 1)
		;
	
	CLEAR_ENABLE;
	CLEAR_RS; CLEAR_RW;

	
	if (on_val == 1)
		data_to_write |= 0x04;
	if (cursor_on == 1)
		data_to_write |= 0x02;	
	if (cursor_blink == 1)
		data_to_write |= 0x01;
	
	PORTA=data_to_write;
	
	TOGGLE_ENABLE;
	
	delay_us (42);
	
} /* end lcd_set_display() */


/* return 1 if LCD is busy */
static int check_bf( void )
{
	int8_t busy = 0;
/* ----------------------------------------------------------------------- */
	
	/* Set Port C for input */
	DDRA=0x00;
	
	/* Turn off pullups */
	PORTA=0x00;
	
	CLEAR_ENABLE;
	delay_us(2);
	CLEAR_RS;
	SET_RW;

	/* Read the busy flag contents */
	SET_ENABLE;
	delay_us(1);
	if (PINA & 0x80)
	{
		busy = 1;
	}
	
	CLEAR_ENABLE;
		
	/* set Port C for output again */
	DDRA=0xFF;

	return busy;
	
} /* end check_bf() */

/*
   lcd_init_instructions ()
   ---------------------
   From the datasheet - re-init the LCD
 */
static void lcd_init_instructions ( void )
{
	delay_ms(15);

	function_mode_set(1,0,0);
	delay_ms(5);

	function_mode_set(1,0,0);
	delay_us(160);

	function_mode_set(1,0,0);
	delay_us(160);	
		
	while (check_bf() == 1)
	{
		function_mode_set(1,0,0);
		STATUS_LED_ON;
		delay_us(10);
		STATUS_LED_OFF;		
		delay_us(140);
	}

	function_mode_set(1,1,0);	
	delay_ms (1);
	
	/* display off */
	lcd_set_display(0,0,0);

	/* display clear */
	lcd_clear ();
	
	/* entry mode set */
	CLEAR_ENABLE;
	CLEAR_RS; CLEAR_RW;
	PORTA=0x06; 
	TOGGLE_ENABLE;
	
	delay_ms (1);
	
	/* display on */
	lcd_set_display(1,0,0);
	
}


/* Initialize PORTA for the data pins and PORTC for other pins */
void lcd_init (void)
{
	/* All port C pins are used for data */
	DDRA=0xFF;
	PORTA=0x00;

	/* RW, RS, Enable pins */
	DDRC |=(_BV(PC3) | _BV(PC4) | _BV(PC5));

	SET_RS;
	CLEAR_RW;
	CLEAR_ENABLE;

	lcd_init_instructions ();
}


