/*********************************************
 * Test of a 20x4 LCD display
 *
 * This program writes the hex digits 0-F on the first line,
 * and Hello World! on the second line of the LCD.
 *
 * PORTA - LCD Data pins DB0-DB7
 * PC5   - Enable
 * PC4   - R/W  L:Data write from MPU to LCM  H:Data read from LCM to MPU
 * PC3   - R/S  L:Instruction code input      H:Data Input
 *
 * Status LED on pin PC0
 *
 * Chip type           : ATMEGA162
 * Clock frequency     : 4.000000 MHz
 *********************************************/

/*************************************************************************
 *			Macros
 *************************************************************************/
/* Set CPU frequency for libc routines and macros */
#define F_CPU 4000000UL  // 4 MHz

#define SET_ENABLE   PORTC |= _BV(PC5);
#define CLEAR_ENABLE PORTC &= ~(_BV(PC5));

//#define TOGGLE_ENABLE SET_ENABLE ; delay_us(1); CLEAR_ENABLE;
#define TOGGLE_ENABLE SET_ENABLE ; delay_us(20); CLEAR_ENABLE;

#define SET_RW       PORTC |= _BV(PC4);
#define CLEAR_RW     PORTC &= ~(_BV(PC4));

#define SET_RS       PORTC |= _BV(PC3);
#define CLEAR_RS     PORTC &= ~(_BV(PC3));

#define STATUS_LED_ON 		PORTC |= _BV(PC0);
#define STATUS_LED_OFF 		PORTC &= ~(_BV(PC0));
/*************************************************************************
 *			Includes
 *************************************************************************/
#include <io.h>
#include <avr/delay.h>


/*************************************************************************
 *			Function Prototypes
 *************************************************************************/

static void delay_ms (uint16_t ms);
static void delay_us (int32_t us);
	
/*************************************************************************
 *			Functions
 *************************************************************************/


/*
  delay_ms
  --------
  Delay up to 2^16 milliseconds using the avr libc delay functions
*/
static void delay_ms (uint16_t ms)
{
	for (;ms>0;ms--)
		_delay_ms(1);
}


/*
  delay_us
  --------
  Delay up to 2^16 microseconds using the avr libc delay functions.
  Accurate to about 10 us.
*/
static void delay_us (int32_t us)
{
	for (;us>0;us--)
		_delay_us(1);
}

/*
  write_lcd_clear 
  ---------------
  Clear the display 
*/
static void write_lcd_clear (void)
{
	/*
	while (check_bf() == 1)
		;
	*/
	CLEAR_ENABLE;
	CLEAR_RS;
	CLEAR_RW;
	delay_us(1);
	
	// Setup the data to send
	PORTA=0x01;

	// Tell the LCD to do it.
	TOGGLE_ENABLE;
	
	// This command takes almost 2ms to complete
	delay_ms(2);
}

/*
  write_lcd_home
  ---------------
  Move cursor to the home position
*/
static void write_lcd_home (void)
{
	/*
	  while (check_bf() == 1)
		;
	*/

	CLEAR_ENABLE;
	CLEAR_RS;
	CLEAR_RW;
	delay_us(1);

	// Setup the data to send
	PORTA=0x02;

	// Tell the LCD to do it.	
	TOGGLE_ENABLE;

	// This command takes almost 2ms to complete
	delay_ms(2);	
	
}

/*
  Write a character at the next position.
*/
static void write_lcd_char(int8_t value)
{
	/*
	while (check_bf() == 1)
		;
	*/
	
	// Set RS high
	// Set RW low to write
	CLEAR_ENABLE;
	SET_RS;
	CLEAR_RW;
	delay_us(1);

	// Setup the data to send
	PORTA=value;

	// Tell the LCD to do it.	
	TOGGLE_ENABLE;
	
	// This operation takes 46us to complete
	delay_us(46);
	
}


/* set_lcd_char_pos
   ----------------
   Sets the position of the cursor for writing the next character.
   The datasheet calls this the DD RAM address.
 */
static void set_lcd_char_pos(int8_t row, int8_t col)
{
	
	// Set RS high
	// Set RW low to write
	CLEAR_ENABLE;
	CLEAR_RS;
	CLEAR_RW;
	delay_us(1);

	// Setup the data to send
	PORTA= 0x80 | (row ? 0x40: 0x00) | col;

	// Tell the LCD to do it.	
	TOGGLE_ENABLE;
	
	// This operation takes 42us to complete
	delay_us(42);	
}

/* function_mode_set
   --------------
   Tells LCD controller about the display
   eight_bit_mode: 0 == 4 it mode 1 == 8 bit mode
   display_lines:  0 == 1 line    1 == 2 lines
   font:           0 == 5x7       1 == 5x10
*/
static void function_mode_set
(
	int8_t eight_bit_mode,
	int8_t  display_lines,
	int8_t font
)
{
	uint8_t data_to_write = 0x20;
	
	CLEAR_ENABLE;
	CLEAR_RS; CLEAR_RW;
	delay_us(1);
	
	if (eight_bit_mode)
		data_to_write |= 0x10;
	if (display_lines)
		data_to_write |= 0x08;
	if (font)
		data_to_write |= 0x04;	
	
	PORTA=data_to_write;
	TOGGLE_ENABLE;
	
	// This operation takes 42us to complete
	delay_us(42);
}

/*
  set_lcd_display
  ---------------
  Write the command to turn the display on or off
  
  on_val:       1 == turn LCD on     0 == turn LCD off
  cursor_on:    1 == turn cursor on  0 == turn cursor off
  cursor_blink: 1 == blink on        0 == blink off
 */
static void set_lcd_display
(
	int8_t on_val,
	int8_t cursor_on,
	int8_t cursor_blink
)
{
	uint8_t data_to_write = 0x08;
/* ------------------------------------------------------------------------- */
	
	/*
	while (check_bf() == 1)
		;
	*/
	
	CLEAR_ENABLE;
	CLEAR_RS; CLEAR_RW;

	
	if (on_val == 1)
		data_to_write |= 0x04;
	if (cursor_on == 1)
		data_to_write |= 0x02;	
	if (cursor_blink == 1)
		data_to_write |= 0x01;
	
	PORTA=data_to_write;
	
	TOGGLE_ENABLE;
	
	delay_us (42);
	
} /* end set_lcd_display() */


/* return 1 if LCD is busy */
static int check_bf( void )
{
	int8_t busy = 0;
/* ----------------------------------------------------------------------- */
	
	/* Set Port C for input */
	DDRA=0x00;
	
	/* Turn off pullups */
	PORTA=0x00;
	
	CLEAR_ENABLE;
	delay_us(2);
	CLEAR_RS;
	SET_RW;

	/* Read the busy flag contents */
	SET_ENABLE;
	delay_us(1);
	if (PINA & 0x80)
	{
		busy = 1;
	}
	
	CLEAR_ENABLE;
		
	/* set Port C for output again */
	DDRA=0xFF;

	return busy;
	
} /* end check_bf() */

/*
   lcd_init_instructions ()
   ---------------------
   From the datasheet - re-init the LCD
 */
static void lcd_init_instructions ( void )
{
	delay_ms(15);

	function_mode_set(1,0,0);
	delay_ms(5);

	function_mode_set(1,0,0);
	delay_us(160);

	function_mode_set(1,0,0);
	delay_us(160);	
		
	while (check_bf() == 1)
	{
		function_mode_set(1,0,0);
		STATUS_LED_ON;
		delay_us(10);
		STATUS_LED_OFF;		
		delay_us(140);
	}

	function_mode_set(1,1,0);	
	delay_ms (1);
	
	/* display off */
	set_lcd_display(0,0,0);

	/* display clear */
	write_lcd_clear ();
	
	/* entry mode set */
	CLEAR_ENABLE;
	CLEAR_RS; CLEAR_RW;
	PORTA=0x06; 
	TOGGLE_ENABLE;
	
	delay_ms (1);
	
	/* display on */
	set_lcd_display(1,0,0);
	
}


/* Initialize PORTA for the data pins and PORTC for other pins */
static void init_lcd (void)
{
	/* All port C pins are used for data */
	DDRA=0xFF;
	PORTA=0x00;

	/* RW, RS, Enable pins */
	DDRC |=(_BV(PC3) | _BV(PC4) | _BV(PC5));

	SET_RS;
	CLEAR_RW;
	CLEAR_ENABLE;

	lcd_init_instructions ();
}

/*
  blink_n_times
  -------------
  Blink the status LED n times (rate is 4 Hz)
 */
static void blink_n_times (int8_t count)
{
	for (; count >=0;count--)
	{
                STATUS_LED_ON;
                delay_ms(100);
                STATUS_LED_OFF;
                delay_ms(150);
	}	
}

/* Main loop of the program
 */
int main(void)
{
	int count ;

	/* Debug LED set to PC0 */
	DDRC|= _BV(PC0);

	/* flash rapidly to show reset */
	for (count = 0; count < 10; count++)
	{
		STATUS_LED_ON;
		delay_ms(25);
		STATUS_LED_OFF;
		delay_ms(75);
	}
	
	/* Initialize LCD Pins */
	init_lcd();

	/* flash slowly 3 times to showwe got past initing the LCD */
	for (count = 0; count < 3; count++)
	{
		STATUS_LED_ON;
		delay_ms(333);
		STATUS_LED_OFF;
		delay_ms(333);
	}
	
	while (1)
	{
		char to_display;
		
                /* Debug led on, pin=0 */
		STATUS_LED_ON;
                delay_ms(500);

		write_lcd_clear();
		write_lcd_home();
		delay_ms(1000);

		// First Line
		for (to_display='0'; to_display <= '9'; to_display++)
		{
			write_lcd_char(to_display);
		}
		for (to_display='A'; to_display <= 'F'; to_display++)
		{
			write_lcd_char(to_display);
		}

		// Second line
		set_lcd_char_pos(1,0);
		
		for (to_display='F'; to_display >= 'A'; to_display--)
		{
			write_lcd_char(to_display);
		}		
		for (to_display='9'; to_display >= '0'; to_display--)
		{
			write_lcd_char(to_display);
		}
		
		delay_ms(5000);
		write_lcd_clear();
		write_lcd_home();

		set_lcd_char_pos(0,5);
		write_lcd_char('H');
		write_lcd_char('e');
		write_lcd_char('l');
		write_lcd_char('l');
		write_lcd_char('o');
		
		
		set_lcd_char_pos(1,5);
		write_lcd_char('W');
		write_lcd_char('o');
		write_lcd_char('r');
		write_lcd_char('l');
		write_lcd_char('d');
		write_lcd_char('!');
		

		STATUS_LED_OFF;
                delay_ms(5000);
	}

	
}

