/*********************************************
* Author: Eric Z. Ayers
* This program is to test the pwm connected to OC1A.
* Trying this with "locked anti-phase PWM" suggested by Dale.
*
* This just tests the H Bridge, running the motor forward for a while,
* and then running it backward for a while.
*
* pin 15 (PD5) Left motor PWM input OC1A
* pin 23 (PC2) Left motor enable
* pin 24 (PC1) Right motor enable
* pin 29 (PE2) Right motor PWM input OC1B
* pin 40 (PC0) Red LED 
*
* Chip type           : ATMEGA162
* Clock frequency     : 4.000000 MHz
*
* Fuse bits of ATMEGA16 factory default
*   Fuse Low Byte      = 0x62
*   Fuse High Byte     = 0x99
*   Fuse Extended Byte = 0xff
*
* Reconfigure fuse bits of ATMEGA162 for 4MHZ external crystal
* Turn off the clock div/8 feature by setting bit 7 to 1
* high nybble of low byte - 0b1010 = A
* low nybble  of low byte - 0b1101 = D
* uisp -dserial=$(PORT) -dprog=avrisp -dpart=$(MCU) --wr_fuse_l=0xAD -wr_fuse_h=0x99
*
*********************************************/

/*************************************************************************
 *			Macros
 *************************************************************************/


/* Set CPU frequency for libc routines and macros */
//#define F_CPU 1000000UL  // 1 MHz
//#define F_CPU 14.7456E6
#define F_CPU 4000000UL  // 4 MHz

#ifndef sbi
#define sbi(_register,_bit_value) (_register) |= _BV(_bit_value)
#endif

#ifndef cbi
#define cbi(_register,_bit_value) (_register) &= ~(_BV(_bit_value))
#endif

#define LEFT_MOTOR_ON 		PORTC |= _BV(PC1)
#define LEFT_MOTOR_OFF 		PORTC &= ~(_BV(PC1))

#define LEFT_MOTOR_FWD_FULL       set_output_oc1a(0xFF)
#define LEFT_MOTOR_FWD_HALF       set_output_oc1a(0xBF)
#define LEFT_MOTOR_STOP           set_output_oc1a(0x7F)
#define LEFT_MOTOR_REV_HALF       set_output_oc1a(0x4F)
#define LEFT_MOTOR_REV_FULL       set_output_oc1a(0x00)

#define RIGHT_MOTOR_ON 		PORTC |= _BV(PC2)
#define RIGHT_MOTOR_OFF 	PORTC &= ~(_BV(PC2))


#define RIGHT_MOTOR_FWD_FULL       set_output_oc1b(0xFF)
#define RIGHT_MOTOR_FWD_HALF       set_output_oc1b(0xBF)
#define RIGHT_MOTOR_STOP           set_output_oc1b(0x7F)
#define RIGHT_MOTOR_REV_HALF       set_output_oc1b(0x4F)
#define RIGHT_MOTOR_REV_FULL       set_output_oc1b(0x00);


/*************************************************************************
 *			Include Files
 *************************************************************************/

#include <io.h>
#include <avr/delay.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

/*************************************************************************
 *			Static Data
 *************************************************************************/


/*************************************************************************
 *			Functions
 *************************************************************************/

/* Delay using the avr libc delay functions */
static void delay_ms (uint16_t ms)
{
	for (;ms>0;ms--)
		_delay_ms(1);
}


static void blink_n_times (int8_t n)
{
	while (n-- > 0)
	{
                /* set output to 5V, LED on */
                PORTC|= _BV(PC0);
                delay_ms(100);

		/* Set pin low, LED off */
                PORTC&= ~_BV(PC0);
                delay_ms(150);
	}
}

static void set_output_oc1a (uint8_t duty)
{
	/* turn on output pin OC1A */
	cbi(TCCR1A, COM1A0); 
	sbi(TCCR1A, COM1A1);
	
	/* Set the duty cycle for in PC1A 0-255 */
	OCR1A = duty;
}

static void set_output_oc1b (uint8_t duty)
{
	/* turn on output pin OC1B */
	cbi(TCCR1A, COM1B0); 
	sbi(TCCR1A, COM1B1);
	
	/* Set the duty cycle for in PC1A 0-255 */
	OCR1B = duty;
}


/*
  enable_pwm
  ----------
  Set timer1 registers to do PWM on ins OC1A and OC1B 
*/
static void enable_pwm(void)
{
	/* Set port D PD5  to output and
	   put it low to start off with. (pwm 50% is stop)
	 */
	DDRD = _BV(PD5);
	PORTD &= ~(_BV(PD5));
	DDRE = _BV(PE2);
	PORTE &= ~(_BV(PE2));
	
	/* Set the Waveform generation mode for timer1
	   ATMEGA32 datasheet page 107

	   Mode 5 8 bit Fast PWM

	   Set OCR1A/OCR1B to be the duty cycle. cycle resets
	   when timer counts to 0xFF
	   
	 */
	cbi(TCCR1B, WGM13);
	sbi(TCCR1B, WGM12);
	cbi(TCCR1A, WGM11);
	sbi(TCCR1A, WGM10);
	
	
	/*
	  I am shooting for approx 100 Hz pwm:

	  clk i/o == 4MHz
	  choose prescaler / 64   clk = 64KHz
	  
	  pwm phase uses 256 clock cycles
	  PWM frequency = 64KHz / 256 = 256 Hz
	  
	 */
	/*
	cbi(TCCR1B,CS12);
	sbi(TCCR1B,CS11);
	sbi(TCCR1B,CS10);
	*/

	/* Try higher frequency... say 2K Hz
	   Use prescaler clk i/o / 8
	 */
	cbi(TCCR1B,CS12);
	sbi(TCCR1B,CS11);
	cbi(TCCR1B,CS10);

	/* Try even higher! no prescaler
	   ... didn't work for some reason...
	 */
	/*
	cbi(TCCR1B,CS12);
	cbi(TCCR1B,CS11);
	cbi(TCCR1B,CS10);
	*/

	
	OCR1A = 0x7F; /* duty cycle for OC1A  50% = full stop */
	OCR1B = 0x7F; /* duty cycle for OC1B  50% = full stop */
	
}

int main(void)
{
	int i;
	
	/* INITIALIZE */

	
	/* Enable PC0 as output for LED
	          PC1 as output for right motor enable
	          PC2 as output for left motor enable		  
	 */
	DDRC |= _BV(PC0) | _BV(PC1) | _BV(PC2);

	/* start with the motors disabled */
	LEFT_MOTOR_OFF;
	RIGHT_MOTOR_OFF;	

	/* Flash the led rapidly to show reset */
	while (i < 20)
	{
		/* off */
		PORTC &= ~_BV(PC0);
		delay_ms(50);
		/* on */
		PORTC |= _BV(PC0);
		delay_ms(5);
		i++;
	}

	
#if 0
	/* Test to just move the left motor forward and backward
	   w/o PWM.
	 */
	DDRD |= _BV(PD5);


	while(1)
	{
		PORTD |= _BV(PD5);
		LEFT_MOTOR_ON;
		delay_ms(4000);
		LEFT_MOTOR_OFF;
		delay_ms(4000);


		PORTD &= ~(_BV(PD5));
		LEFT_MOTOR_ON;
		delay_ms(4000);
		LEFT_MOTOR_OFF;
		delay_ms(4000);		
	}
#endif
	
	/* Init PWM registers
	 */
	enable_pwm();


	/* Loop forever */
	while (1)
	{

		delay_ms(1000);
		blink_n_times(1);
		delay_ms(1000);
		
		/* Test the left motor */
		LEFT_MOTOR_ON;
		
		LEFT_MOTOR_REV_FULL;
		delay_ms(4000);
		blink_n_times(1);
		LEFT_MOTOR_REV_HALF;
		delay_ms(4000);
		blink_n_times(2);
		LEFT_MOTOR_STOP;
		delay_ms(4000);
		blink_n_times(3);		
		LEFT_MOTOR_FWD_HALF;
		delay_ms(4000);
		blink_n_times(4);		
		LEFT_MOTOR_FWD_FULL;
		delay_ms(4000);
		blink_n_times(3);		
		LEFT_MOTOR_FWD_HALF;
		delay_ms(4000);
		blink_n_times(2);		
		LEFT_MOTOR_STOP;
		delay_ms(4000);
		blink_n_times(1);		
		LEFT_MOTOR_REV_HALF;
		delay_ms(4000);
		LEFT_MOTOR_REV_FULL;
		delay_ms(4000);

		LEFT_MOTOR_OFF;

		delay_ms(4000);
		blink_n_times(2);
		delay_ms(4000);
		
		/* Test the right motor */
		RIGHT_MOTOR_ON;
		
		RIGHT_MOTOR_REV_FULL;
		delay_ms(4000);
		blink_n_times(1);
		RIGHT_MOTOR_REV_HALF;
		delay_ms(4000);
		blink_n_times(2);
		RIGHT_MOTOR_STOP;
		delay_ms(4000);
		blink_n_times(3);		
		RIGHT_MOTOR_FWD_HALF;
		delay_ms(4000);
		blink_n_times(4);		
		RIGHT_MOTOR_FWD_FULL;
		delay_ms(4000);
		blink_n_times(3);		
		RIGHT_MOTOR_FWD_HALF;
		delay_ms(4000);
		blink_n_times(2);		
		RIGHT_MOTOR_STOP;
		delay_ms(4000);
		blink_n_times(1);		
		RIGHT_MOTOR_REV_HALF;
		delay_ms(4000);
		RIGHT_MOTOR_REV_FULL;
		delay_ms(4000);

		RIGHT_MOTOR_OFF;
		
		delay_ms(1000);
		blink_n_times(3);
		delay_ms(1000);
		
		/* Run both at the same time */
		RIGHT_MOTOR_ON;
		LEFT_MOTOR_ON;		
		
		RIGHT_MOTOR_REV_FULL;
		LEFT_MOTOR_REV_FULL;
		delay_ms(4000);
		blink_n_times(1);
		RIGHT_MOTOR_REV_HALF;
		LEFT_MOTOR_REV_HALF;		
		delay_ms(4000);
		blink_n_times(2);
		RIGHT_MOTOR_STOP;
		LEFT_MOTOR_STOP;		
		delay_ms(4000);
		blink_n_times(3);		
		RIGHT_MOTOR_FWD_HALF;
		LEFT_MOTOR_FWD_HALF;		
		delay_ms(4000);
		blink_n_times(4);		
		RIGHT_MOTOR_FWD_FULL;
		LEFT_MOTOR_FWD_FULL;		
		delay_ms(4000);
		blink_n_times(3);		
		RIGHT_MOTOR_FWD_HALF;
		LEFT_MOTOR_FWD_HALF;		
		delay_ms(4000);
		blink_n_times(2);		
		RIGHT_MOTOR_STOP;
		LEFT_MOTOR_STOP;		
		delay_ms(4000);
		blink_n_times(1);		
		RIGHT_MOTOR_REV_HALF;
		LEFT_MOTOR_REV_HALF;		
		delay_ms(4000);
		RIGHT_MOTOR_REV_FULL;
		LEFT_MOTOR_REV_FULL;		
		delay_ms(4000);

		RIGHT_MOTOR_OFF;
		LEFT_MOTOR_OFF;		

		delay_ms(1000);
		blink_n_times(4);
		delay_ms(10000);

	}
}

