/*********************************************
* Author: Eric Z. Ayers
* This program is to test the pwm connected to OC1A.
* Trying this with "locked anti-phase PWM" suggested by Dale.
*
* This just tests the H Bridge, running the motor forward for a while,
* and then running it backward for a while.
*
* pin 15 (PD5) Left motor PWM input OC1A
* pin 23 (PC2) Left motor enable         
* pin 40 (PC0) Red LED 
*
* Chip type           : ATMEGA162
* Clock frequency     : 4.000000 MHz
*
* Fuse bits of ATMEGA16 factory default
*   Fuse Low Byte      = 0x62
*   Fuse High Byte     = 0x99
*   Fuse Extended Byte = 0xff
*
* Reconfigure fuse bits of ATMEGA162 for 4MHZ external crystal
* Turn off the clock div/8 feature by setting bit 7 to 1
* high nybble of low byte - 0b1010 = A
* low nybble  of low byte - 0b1101 = D
* uisp -dserial=$(PORT) -dprog=avrisp -dpart=$(MCU) --wr_fuse_l=0xAD -wr_fuse_h=0x99
*
*********************************************/

/*************************************************************************
 *			Macros
 *************************************************************************/


/* Set CPU frequency for libc routines and macros */
//#define F_CPU 1000000UL  // 1 MHz
//#define F_CPU 14.7456E6
#define F_CPU 4000000UL  // 4 MHz

#ifndef sbi
#define sbi(_register,_bit_value) (_register) |= _BV(_bit_value)
#endif

#ifndef cbi
#define cbi(_register,_bit_value) (_register) &= ~(_BV(_bit_value))
#endif

#define RIGHT_MOTOR_ON 		PORTC |= _BV(PC2)
#define RIGHT_MOTOR_OFF 	PORTC &= ~(_BV(PC2))

#define LEFT_MOTOR_ON 		PORTC |= _BV(PC1)
#define LEFT_MOTOR_OFF 		PORTC &= ~(_BV(PC1))

/*************************************************************************
 *			Include Files
 *************************************************************************/

#include <io.h>
#include <avr/delay.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

/*************************************************************************
 *			Static Data
 *************************************************************************/


/*************************************************************************
 *			Functions
 *************************************************************************/

/* Delay using the avr libc delay functions */
static void delay_ms (uint16_t ms)
{
	for (;ms>0;ms--)
		_delay_ms(1);
}


static void blink_n_times (int8_t n)
{
	while (n-- > 0)
	{
                /* set output to 5V, LED on */
                PORTC|= _BV(PC0);
                delay_ms(500);

		/* Set pin low, LED off */
                PORTC&= ~_BV(PC0);
                delay_ms(500);
	}
}

static void set_output_oc1a (uint8_t duty)
{
	// Set pin OC1B low
	OCR1A = 0;
	
	/* turn on output pin OC1A */
	cbi(TCCR1A, COM1A0); 
	sbi(TCCR1A, COM1A1);
	
	/* Set the duty cycle for in PC1A 0-255 */
	OCR1A = duty;
}


/*
  enable_pwm
  ----------
  Set timer1 registers to do PWM on ins OC1A and OC1B 
*/
static void enable_pwm(void)
{
	/* Set port D PD5  to output and
	   put it low to start off with. (pwm 50% is stop)
	 */
	DDRD = _BV(PD5);
	PORTD &= ~(_BV(PD5));


	/* Set the Waveform generation mode for timer1
	   ATMEGA32 datasheet page 107

	   Mode 5 8 bit Fast PWM

	   Set OCR1A/OCR1B to be the duty cycle. cycle resets
	   when timer counts to 0xFF
	   
	 */
	cbi(TCCR1B, WGM13);
	sbi(TCCR1B, WGM12);
	cbi(TCCR1A, WGM11);
	sbi(TCCR1A, WGM10);
	
	
	/*
	  I am shooting for approx 100 Hz pwm:

	  clk i/o == 4MHz
	  choose prescaler / 64   clk = 64KHz
	  
	  pwm phase uses 256 clock cycles
	  PWM frequency = 64KHz / 256 = 256 Hz
	  
	 */
	/*
	cbi(TCCR1B,CS12);
	sbi(TCCR1B,CS11);
	sbi(TCCR1B,CS10);
	*/

	/* Try higher frequency... say 2K Hz
	   Use prescaler clk i/o / 8
	 */
	cbi(TCCR1B,CS12);
	sbi(TCCR1B,CS11);
	cbi(TCCR1B,CS10);

	/* Try even higher! no prescaler
	   ... didn't work for some reason...
	 */
	/*
	cbi(TCCR1B,CS12);
	cbi(TCCR1B,CS11);
	cbi(TCCR1B,CS10);
	*/

	
	OCR1A = 0x00; /* duty cycle for OC1A */
}

int main(void)
{
	/* INITIALIZE */

	
	/* Enable PC0 as output for LED
	          PC1 as output for right motor enable
	          PC2 as output for left motor enable		  
	 */
	DDRC |= _BV(PC0) | _BV(PC1) | _BV(PC2);

#if 0
	/* Test to just move the left motor forward and backward
	   w/o PWM.
	 */
	DDRD |= _BV(PD5);


	while(1)
	{
		PORTD |= _BV(PD5);
		LEFT_MOTOR_ON;
		delay_ms(4000);
		LEFT_MOTOR_OFF;
		delay_ms(4000);


		PORTD &= ~(_BV(PD5));
		LEFT_MOTOR_ON;
		delay_ms(4000);
		LEFT_MOTOR_OFF;
		delay_ms(4000);		
	}
#endif
	
	/* Init PWM registers
	 */
	enable_pwm();


	/* Loop forever */
	while (1)
	{

		delay_ms(1000);

		/* Turn on the left motor */
		LEFT_MOTOR_ON;
		
		/* Turn on the PWM for the specified duty (0-256) */
		set_output_oc1a(0);
		delay_ms(4000);
		blink_n_times(1);
		set_output_oc1a(64);
		delay_ms(4000);
		blink_n_times(2);
		set_output_oc1a(128);
		delay_ms(4000);
		blink_n_times(3);		
		set_output_oc1a(192);
		delay_ms(4000);
		blink_n_times(4);		
		set_output_oc1a(0xFF);
		delay_ms(4000);
		blink_n_times(3);		
		set_output_oc1a(192);
		delay_ms(4000);
		blink_n_times(2);		
		set_output_oc1a(128);
		delay_ms(4000);
		blink_n_times(1);		
		set_output_oc1a(64);
		delay_ms(4000);
		set_output_oc1a(0);
		delay_ms(4000);

		LEFT_MOTOR_OFF;
		
		blink_n_times(2);
		delay_ms(1000);
		
	}
}

