/*************************************************************************
* Author: Eric Z. Ayers
* This program is to test the pwm connected to OC1A.
* Trying this with "locked anti-phase PWM" suggested by Dale.
*
* This just tests the H Bridge, running the motor forward for a while,
* and then running it backward for a while.
*
* pin 24 (PC1) Right motor enable
* pin 23 (PC2) Left motor enable
* pin 29 (PE2) Right motor PWM input OC1B
* pin 40 (PC0) Red LED
* pin 10 (PD0, TXD0) RS232 Transmit (Debugging 38400 bps)
* pin 11 (PD1, RXD0) RS232 Receive (Debugging 38400 bps)
* pin 12 (PD2, INT0) Left wheel encoder input
* pin 13 (PD3, INT1) Right wheel encoder input
* pin 15 (PD5) Left motor PWM input OC1A
*
* Chip type           : ATMEGA162
* Clock frequency     : 4.000000 MHz
*
* Fuse bits of ATMEGA16 factory default
*   Fuse Low Byte      = 0x62
*   Fuse High Byte     = 0x99
*   Fuse Extended Byte = 0xff
*
* Reconfigure fuse bits of ATMEGA162 for 4MHZ external crystal
* Turn off the clock div/8 feature by setting bit 7 to 1
* high nybble of low byte - 0b1010 = A
* low nybble  of low byte - 0b1101 = D
* uisp -dserial=$(PORT) -dprog=avrisp -dpart=$(MCU) --wr_fuse_l=0xAD -wr_fuse_h=0x99
*
**************************************************************************/

/*************************************************************************
 *			Macros
 *************************************************************************/


/* Set CPU frequency for libc routines and macros */
//#define F_CPU 1000000UL  // 1 MHz
//#define F_CPU 14.7456E6
#define F_CPU 4000000UL  // 4 MHz

#ifndef sbi
#define sbi(_register,_bit_value) (_register) |= _BV(_bit_value)
#endif

#ifndef cbi
#define cbi(_register,_bit_value) (_register) &= ~(_BV(_bit_value))
#endif

#define LEFT_MOTOR_ON 		PORTC |= _BV(PC1)
#define LEFT_MOTOR_OFF 		PORTC &= ~(_BV(PC1))

/* Pass these macro a number between 0 and 0x7F (127) to
   move forward at a speed where 0 is stop and 0x7F is full forward.
 */
#define RIGHT_MOTOR_FWD(arg)       set_output_oc1b(0x7F - (arg));
#define LEFT_MOTOR_FWD(arg)       set_output_oc1a(0x7F - (arg));


#define LEFT_MOTOR_FWD_FULL       set_output_oc1a(0x00)
#define LEFT_MOTOR_FWD_HALF       set_output_oc1a(0x4F)
#define LEFT_MOTOR_STOP           set_output_oc1a(0x7F)
#define LEFT_MOTOR_REV_HALF       set_output_oc1a(0xBF)
#define LEFT_MOTOR_REV_FULL       set_output_oc1a(0xFF)

#define RIGHT_MOTOR_ON 		PORTC |= _BV(PC2)
#define RIGHT_MOTOR_OFF 	PORTC &= ~(_BV(PC2))

#define RIGHT_MOTOR_FWD_FULL       set_output_oc1b(0x00);
#define RIGHT_MOTOR_FWD_HALF       set_output_oc1b(0x4F)
#define RIGHT_MOTOR_STOP           set_output_oc1b(0x7F)
#define RIGHT_MOTOR_REV_HALF       set_output_oc1b(0xBF)
#define RIGHT_MOTOR_REV_FULL       set_output_oc1b(0xFF)

#define STATUS_LED_ON 		PORTC |= _BV(PC0)
#define STATUS_LED_OFF 		PORTC &= ~(_BV(PC0))

/*************************************************************************
 *			Include Files
 *************************************************************************/

#include <io.h>
#include <avr/delay.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

/*************************************************************************
 *			Static Data
 *************************************************************************/
static unsigned char led_state = 0;
static uint16_t left_wheel_count = 0;
static uint16_t right_wheel_count = 0;

/*************************************************************************
 *			Functions
 *************************************************************************/

/* Delay using the avr libc delay functions */
static void delay_ms (uint16_t ms)
{
	for (;ms>0;ms--)
		_delay_ms(1);
}


/* Blink the LED 'n' number of times. (one cycle takes 250ms) */
static void blink_n_times (int8_t n)
{
	while (n-- > 0)
	{
                /* set output to 5V, LED on */
                STATUS_LED_ON;
                delay_ms(100);

		/* Set pin low, LED off */
		STATUS_LED_OFF;
                delay_ms(150);
	}
}

/* Left wheel encoder changed
   INT0 == Port PD2
 */
SIGNAL(SIG_INTERRUPT0)
{
	left_wheel_count++;
	
	if (led_state == 0)
	{
		/* turn off the LED */
		STATUS_LED_OFF;
		led_state = 1;
	}
	else
	{
		/* turn on the LED */
		STATUS_LED_ON;
		led_state = 0;
	}
}


/* Right wheel encoder changed */
SIGNAL(SIG_INTERRUPT1)
{
	right_wheel_count++;
	
	if (led_state == 0)
	{
		/* turn off the LED */
		STATUS_LED_OFF;
		led_state = 1;		
	}
	else
	{
		/* turn on the LED */
		STATUS_LED_ON;
		led_state = 0;		
	}	
}


/*
  setup_serial_port0
  ------------------
  
  Setup the USART port as an asyncronous serial port running
  at 38400 bps, 8,N,1.

  The datasheet says there are 2 USARTs, but the datasheet only
  specifies one set of registers for these settings in the details.
  The intro notes that you use a '0' or '1' as the pentultimate character
  in the register name to indicate which port you mean:

  UCSRA --> UCSR0A and UCSR1A

  Ditto for the bit constants (why?)
  
 */
static void setup_serial_port0 ( void )
{
	/* Table on page 191 for setting baud rates */
	/* 38400 bps  4MHz 
	   UBRR = 12, U2X = 1
	*/
	UBRR0H = 0x80;
	UBRR0L = 0x0C;
	UCSR0A |= _BV(U2X0);  /* Double the xfer rate selected in UBBR */

	/* Turn on transmit & receive mode */
	UCSR0B = _BV(RXEN0) | _BV(TXEN0);
	
	/* Set to Asyncronous 8 bit word, No parity, 1 stop bit */
	UCSR0C = _BV(URSEL0) | _BV(UCSZ01) | _BV(UCSZ00);
	
}

/* Write a character to the serial port 0 */
static void write_serial_port0 (uint8_t data)
{
	STATUS_LED_ON;
	delay_ms(5);
	loop_until_bit_is_set (UCSR0A, UDRE0);
	UDR0 = data;
	
	STATUS_LED_OFF;
	delay_ms(20);
}


/* Wait for a character to become available and read a character
   from the serial port.  If you don't want this function to block,
   then check the RCX0 bit of UCSR0A before calling.
*/ 
static uint8_t read_serial_port0 ( void )
{
	loop_until_bit_is_set (UCSR0A, RXC0);
	return UDR0;
}

int main(void)
{
	int i = 0;
	
	/* INITIALIZE */

	
	/* Enable PC0 as output for LED
	          PC1 as output for right motor enable
	          PC2 as output for left motor enable		  
	 */
	DDRC |= _BV(PC0) | _BV(PC1) | _BV(PC2);


	/* Flash the led rapidly to show reset */
	while (i < 20)
	{
		/* off */
		STATUS_LED_OFF;
		delay_ms(50);
		/* on */
		STATUS_LED_ON;
		delay_ms(5);
		i++;
	}
	
	STATUS_LED_OFF;
	
	/* Enable external interrupts for wheel encoders */
	enable_external_int(_BV(INT0)|_BV(INT1));

	/* An interrupt should occur whenever a change from low to
	   high or high to low occurs
	*/
	MCUCR = _BV(ISC10) | _BV(ISC00);
	
	/* Enable system wide interrupt flag */
	sei();


	/* Setup the serial port for rs232 communication*/
	setup_serial_port0();

	/* delay before main loop  to allow user to get ready */
	delay_ms(4000);


	/* Write out the letters A-Z, then wait for a character and
	   output that character plus that man instances of the character '+'
	*/
	while (1)
	{
		uint8_t read_data;

		blink_n_times(1);
		delay_ms(1000);
		
		for (i=0;i<26;i++)
		{
			write_serial_port0('A' + i);
		}

		/* Write CR/LF */
		write_serial_port0('\r');
		write_serial_port0('\n');

		blink_n_times(2);
		delay_ms(1000);
		
		/* Write a prompt character to the user */
		write_serial_port0('>');
		
		/* Wait for a character */
		read_data = read_serial_port0();

		/* Echo the character back to the user */
		write_serial_port0(read_data);
		write_serial_port0('\r');
		write_serial_port0('\n');

		blink_n_times(3);
		delay_ms(1000);

		/* Output a stream of '+' characters to represent the
		   ASCII value
		*/
		for (i=0;i<read_data;i++)
		{
			write_serial_port0('+');
		}
		
		/* Write CR/LF */
		write_serial_port0('\r');
		write_serial_port0('\n');
		delay_ms(1000);
		
	} /* end loop forever */
	
}
