/*********************************************
* Author: Eric Z. Ayers
* This program is to test the pwm connected to OC1A.
*
* This just tests the H Bridge, running the motor forward for a while,
* and then running it backward for a while.
*
* pin 24 (PC1) Right motor enable
* pin 23 (PC2) Left motor enable
* pin 29 (PE2) Right motor PWM input OC1B
* pin 40 (PC0) Red LED
* pin 10 (PD0, TXD0) RS232 Transmit (Debugging 38400 bps)
* pin 11 (PD1, RXD0) RS232 Receive (Debugging 38400 bps)
* pin 12 (PD2, INT0) Left wheel encoder input
* pin 13 (PD3, INT1) Right wheel encoder input
* pin 15 (PD5) Left motor PWM input OC1A
*
* Chip type           : ATMEGA162
* Clock frequency     : 4.000000 MHz
*
* Fuse bits of ATMEGA16 factory default
*   Fuse Low Byte      = 0x62
*   Fuse High Byte     = 0x99
*   Fuse Extended Byte = 0xff
*
* Reconfigure fuse bits of ATMEGA162 for 4MHZ external crystal
* Turn off the clock div/8 feature by setting bit 7 to 1
* high nybble of low byte - 0b1010 = A
* low nybble  of low byte - 0b1101 = D
* uisp -dserial=$(PORT) -dprog=avrisp -dpart=$(MCU) --wr_fuse_l=0xAD -wr_fuse_h=0x99
*
*********************************************/

/*************************************************************************
 *			Macros
 *************************************************************************/


/* Set CPU frequency for libc routines and macros */
//#define F_CPU 1000000UL  // 1 MHz
//#define F_CPU 14.7456E6
#define F_CPU 4000000UL  // 4 MHz

#ifndef sbi
#define sbi(_register,_bit_value) (_register) |= _BV(_bit_value)
#endif

#ifndef cbi
#define cbi(_register,_bit_value) (_register) &= ~(_BV(_bit_value))
#endif

#define LEFT_MOTOR_ON 		PORTC |= _BV(PC1)
#define LEFT_MOTOR_OFF 		PORTC &= ~(_BV(PC1))

/* Pass these macro a number between 0 and 0x7F (127) to
   move forward at a speed where 0 is stop and 0x7F is full forward.
 */
#define RIGHT_MOTOR_FWD(arg)       set_output_oc1b(0x7F - (arg));
#define LEFT_MOTOR_FWD(arg)       set_output_oc1a(0x7F - (arg));


#define LEFT_MOTOR_FWD_FULL       set_output_oc1a(0x00)
#define LEFT_MOTOR_FWD_HALF       set_output_oc1a(0x4F)
#define LEFT_MOTOR_STOP           set_output_oc1a(0x7F)
#define LEFT_MOTOR_REV_HALF       set_output_oc1a(0xBF)
#define LEFT_MOTOR_REV_FULL       set_output_oc1a(0xFF)

#define RIGHT_MOTOR_ON 		PORTC |= _BV(PC2)
#define RIGHT_MOTOR_OFF 	PORTC &= ~(_BV(PC2))

#define RIGHT_MOTOR_FWD_FULL       set_output_oc1b(0x00);
#define RIGHT_MOTOR_FWD_HALF       set_output_oc1b(0x4F)
#define RIGHT_MOTOR_STOP           set_output_oc1b(0x7F)
#define RIGHT_MOTOR_REV_HALF       set_output_oc1b(0xBF)
#define RIGHT_MOTOR_REV_FULL       set_output_oc1b(0xFF)

#define STATUS_LED_ON 		PORTC |= _BV(PC0)
#define STATUS_LED_OFF 		PORTC &= ~(_BV(PC0))

/*************************************************************************
 *			Include Files
 *************************************************************************/

#include <io.h>
#include <avr/delay.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

/*************************************************************************
 *			Static Data
 *************************************************************************/
static unsigned char led_state = 0;

/*************************************************************************
 *			Functions
 *************************************************************************/

/* Delay using the avr libc delay functions */
static void delay_ms (uint16_t ms)
{
	for (;ms>0;ms--)
		_delay_ms(1);
}


static void blink_n_times (int8_t n)
{
	while (n-- > 0)
	{
                /* set output to 5V, LED on */
                STATUS_LED_ON;
                delay_ms(100);

		/* Set pin low, LED off */
		STATUS_LED_OFF;
                delay_ms(150);
	}
}

/* Left wheel encoder changed
   INT0 == Port PD2
 */
SIGNAL(SIG_INTERRUPT0)
{
	if (led_state == 0)
	{
		/* turn off the LED */
		STATUS_LED_OFF;
		led_state = 1;
	}
	else
	{
		/* turn on the LED */
		STATUS_LED_ON;
		led_state = 0;
	}
}


/* Right wheel encoder changed */
SIGNAL(SIG_INTERRUPT1)
{
	if (led_state == 0)
	{
		/* turn off the LED */
		STATUS_LED_OFF;
		led_state = 1;		
	}
	else
	{
		/* turn on the LED */
		STATUS_LED_ON;
		led_state = 0;		
	}	
}


int main(void)
{
	int i = 0;
	
	/* INITIALIZE */

	
	/* Enable PC0 as output for LED
	          PC1 as output for right motor enable
	          PC2 as output for left motor enable		  
	 */
	DDRC |= _BV(PC0) | _BV(PC1) | _BV(PC2);


	/* Flash the led rapidly to show reset */
	while (i < 20)
	{
		/* off */
		STATUS_LED_OFF;
		delay_ms(50);
		/* on */
		STATUS_LED_ON;
		delay_ms(5);
		i++;
	}

	/* Enable external interrupts for wheel encoders */
	enable_external_int(_BV(INT0)|_BV(INT1));

	/* An interrupt should occur whenever a change from low to
	   high or high to low occurs
	*/
	MCUCR = _BV(ISC10) | _BV(ISC00);
	
	/* Enable system wide interrupt flag */
	sei();

	while (1)
	{
		/* wait for interrupts to go off */
	}
}
