/*********************************************
* Author: Eric Z. Ayers
* This program is to test LCD attached to the second serial port to
* print the encoder output.
*
* pin 12 (PD2, INT0) Left wheel encoder input
* pin 13 (PD3, INT1) Right wheel encoder input
* pin    (PD0, RXD0) serial port 0 receive  - to computer for debugging
* pin    (PD1, TXD0) serial port 0 transmit
* pin    (PB2, RXD1) serial port 1 receive  - to LCD for debugging
* pin    (PB3, TXD1) serial port 1 transmit
*
* Chip type           : ATMEGA162
* Clock frequency     : 4.000000 MHz
*
* Fuse bits of ATMEGA16 factory default
*   Fuse Low Byte      = 0x62
*   Fuse High Byte     = 0x99
*   Fuse Extended Byte = 0xff
*
* Reconfigure fuse bits of ATMEGA162 for 4MHZ external crystal
* Turn off the clock div/8 feature by setting bit 7 to 1
* high nybble of low byte - 0b1010 = A
* low nybble  of low byte - 0b1101 = D
* uisp -dserial=$(PORT) -dprog=avrisp -dpart=$(MCU) --wr_fuse_l=0xAD -wr_fuse_h=0x99
*
*********************************************/

/*************************************************************************
 *			Macros
 *************************************************************************/


/* Set CPU frequency for libc routines and macros */
//#define F_CPU 1000000UL  // 1 MHz
//#define F_CPU 14.7456E6
#define F_CPU 4000000UL  // 4 MHz

#ifndef sbi
#define sbi(_register,_bit_value) (_register) |= _BV(_bit_value)
#endif

#ifndef cbi
#define cbi(_register,_bit_value) (_register) &= ~(_BV(_bit_value))
#endif

#define LEFT_MOTOR_ON 		PORTC |= _BV(PC1)
#define LEFT_MOTOR_OFF 		PORTC &= ~(_BV(PC1))

/* Pass these macro a number between 0 and 0x7F (127) to
   move forward at a speed where 0 is stop and 0x7F is full forward.
 */
#define RIGHT_MOTOR_FWD(arg)       set_output_oc1b(0x7F - (arg));
#define LEFT_MOTOR_FWD(arg)       set_output_oc1a(0x7F - (arg));


#define LEFT_MOTOR_FWD_FULL       set_output_oc1a(0x00)
#define LEFT_MOTOR_FWD_HALF       set_output_oc1a(0x4F)
#define LEFT_MOTOR_STOP           set_output_oc1a(0x7F)
#define LEFT_MOTOR_REV_HALF       set_output_oc1a(0xBF)
#define LEFT_MOTOR_REV_FULL       set_output_oc1a(0xFF)

#define RIGHT_MOTOR_ON 		PORTC |= _BV(PC2)
#define RIGHT_MOTOR_OFF 	PORTC &= ~(_BV(PC2))

#define RIGHT_MOTOR_FWD_FULL       set_output_oc1b(0x00);
#define RIGHT_MOTOR_FWD_HALF       set_output_oc1b(0x4F)
#define RIGHT_MOTOR_STOP           set_output_oc1b(0x7F)
#define RIGHT_MOTOR_REV_HALF       set_output_oc1b(0xBF)
#define RIGHT_MOTOR_REV_FULL       set_output_oc1b(0xFF)

#define STATUS_LED_ON 		PORTC |= _BV(PC0)
#define STATUS_LED_OFF 		PORTC &= ~(_BV(PC0))

/*************************************************************************
 *			Include Files
 *************************************************************************/

#include <io.h>
#include <avr/delay.h>
#include <avr/interrupt.h>
#include <avr/signal.h>

#include <stdlib.h>
#include <string.h>

#include "lcd_api.h"

/*************************************************************************
 *			Static Data
 *************************************************************************/
static unsigned char gbl_led_state = 0;
static uint32_t      gbl_right_encoder_tics = 0;
static uint32_t      gbl_left_encoder_tics = 0;

/*************************************************************************
 *			Functions
 *************************************************************************/

/* Delay using the avr libc delay functions */
static void delay_ms (uint16_t ms)
{
	for (;ms>0;ms--)
		_delay_ms(1);
}


static void blink_n_times (int8_t n)
{
	while (n-- > 0)
	{
                /* set output to 5V, LED on */
                STATUS_LED_ON;
                delay_ms(100);

		/* Set pin low, LED off */
		STATUS_LED_OFF;
                delay_ms(150);
	}
}

/* Left wheel encoder changed
   INT0 == Port PD2
 */
SIGNAL(SIG_INTERRUPT0)
{
	gbl_left_encoder_tics++;			
	if (gbl_led_state == 0)
	{
		/* turn off the LED */
		STATUS_LED_OFF;
		gbl_led_state = 1;
	}
	else
	{
		/* turn on the LED */
		STATUS_LED_ON;
		gbl_led_state = 0;
	}
}


/* Right wheel encoder changed */
SIGNAL(SIG_INTERRUPT1)
{
	gbl_right_encoder_tics++;
	
	if (gbl_led_state == 0)
	{
		/* turn off the LED */
		STATUS_LED_OFF;
		gbl_led_state = 1;		
	}
	else
	{
		/* turn on the LED */
		STATUS_LED_ON;
		gbl_led_state = 0;		
	}	
}

/*****************************************************************************
 * Serial port stuff
 */
/*
  setup_serial_port0
  ------------------
  
  Setup the USART port as an asyncronous serial port running
  at 38400 bps, 8,N,1.

  The datasheet says there are 2 USARTs, but the datasheet only
  specifies one set of registers for these settings in the details.
  The intro notes that you use a '0' or '1' as the pentultimate character
  in the register name to indicate which port you mean:

  UCSRA --> UCSR0A and UCSR1A

  Ditto for the bit constants (why?)
  
 */
static void setup_serial_port0 ( void )
{
	/* Table on page 191 for setting baud rates */
	/* 38400 bps  4MHz 
	   UBRR = 12, U2X = 1
	*/
	UBRR0H = 0x80;
	UBRR0L = 0x0C;
	UCSR0A |= _BV(U2X0);  /* Double the xfer rate selected in UBBR */

	/* Turn on transmit & receive mode */
	UCSR0B = _BV(RXEN0) | _BV(TXEN0);
	
	/* Set to Asyncronous 8 bit word, No parity, 1 stop bit */
	UCSR0C = _BV(URSEL0) | _BV(UCSZ01) | _BV(UCSZ00);
	
}

/* Write a character to the serial port 0 */
static void write_serial_port0 (uint8_t data)
{
	STATUS_LED_ON;
	delay_ms(5);
	loop_until_bit_is_set (UCSR0A, UDRE0);
	UDR0 = data;
	
	STATUS_LED_OFF;
	delay_ms(20);
}


/* Wait for a character to become available and read a character
   from the serial port.  If you don't want this function to block,
   then check the RCX0 bit of UCSR0A before calling.
*/ 
static uint8_t read_serial_port0 ( void )
{
	loop_until_bit_is_set (UCSR0A, RXC0);
	return UDR0;
}



/*
  setup_serial_port1
  ------------------
  
  Setup the second USART port as an asyncronous serial port running
  at 38400 bps, 8,N,1.
  
 */
static void setup_serial_port1 ( void )
{
	/* Table on page 191 for setting baud rates */
	/* 38400 bps  4MHz 
	   UBRR = 12, U2X = 1
	*/
	UBRR1H = 0x80;
	UBRR1L = 0x0C;
	UCSR1A |= _BV(U2X1);  /* Double the xfer rate selected in UBBR */

	/* Turn on transmit & receive mode */
	UCSR1B = _BV(RXEN1) | _BV(TXEN1);
	
	/* Set to Asyncronous 8 bit word, No parity, 1 stop bit */
	UCSR1C = _BV(URSEL1) | _BV(UCSZ11) | _BV(UCSZ10);
	
}

/* Write a character to the serial port 0 */
static void write_serial_port1 (uint8_t data)
{
	STATUS_LED_ON;
	delay_ms(5);
	loop_until_bit_is_set (UCSR1A, UDRE1);
	UDR0 = data;
	
	STATUS_LED_OFF;
	delay_ms(20);
}

/* Wait for a character to become available and read a character
   from the serial port.  If you don't want this function to block,
   then check the RCX0 bit of UCSR0A before calling.
*/ 
static uint8_t read_serial_port1 ( void )
{
	loop_until_bit_is_set (UCSR1A, RXC1);
	return UDR1;
}


static void write_serial_port_string(int8_t port_num, char *data)
{
	char *p;
	for (p=data;p;p++)
	{
		if (port_num == 0)
			write_serial_port0 (*p);
		else
			write_serial_port1 (*p);
	}
}

/* Writes the specified value as a string across the serial port
 */
static void write_serial_port_uint(int8_t port_num, uint32_t data)
{
#define BUFLEN 9
	char buf[9];
	int8_t buf_index = BUFLEN -1;

	if (data == 0)
	{
		if (port_num == 0)
			write_serial_port0 ('0');
		else
			write_serial_port1 ('0');
		return;
	}

	memset(buf, 0, sizeof(buf));
	
	/* Copy the integer into the string buffer.
	   This creates a right justified string value
	 */
	while (data >0 && buf_index >=0)
	{
		buf[buf_index] = data %10;
		data/=10;
	}

	/* Now output data to the serial port */
	for (buf_index=0;buf_index<BUFLEN;buf_index++)
	{
		if (buf[buf_index == 0])
			continue;
		
		if (port_num == 0)
			write_serial_port0 (buf[buf_index]);
		else
			write_serial_port1 (buf[buf_index]);
	}
	
#undef BUFLEN		
} /* end write_serial_port_uint() */

int main(void)
{
	int i = 0;
	uint16_t loops = 0;
	
	/* INITIALIZE */

	
	/* Enable PC0 as output for LED
	          PC1 as output for right motor enable
	          PC2 as output for left motor enable		  
	 */
	DDRC |= _BV(PC0) | _BV(PC1) | _BV(PC2);

	/* Set PB3 and PD1 for output for serial port transmit */
	DDRB |= _BV(PB3);
	DDRD |= _BV(PD1);
	
	/* Flash the led rapidly to show reset */
	while (i < 20)
	{
		/* off */
		STATUS_LED_OFF;
		delay_ms(50);
		/* on */
		STATUS_LED_ON;
		delay_ms(5);
		i++;
	}

	/* Enable external interrupts for wheel encoders */
	enable_external_int(_BV(INT0)|_BV(INT1));

	/* An interrupt should occur whenever a change from low to
	   high or high to low occurs
	*/
	MCUCR = _BV(ISC10) | _BV(ISC00);
	
	/* Enable system wide interrupt flag */
	sei();

	setup_serial_port0();
	setup_serial_port1();

	/* Initialize the LCD directly connected to portA  */
	lcd_init();
	
	while (1)
	{
		loops++;
		
		/* wait for interrupts to go off */
		delay_ms(1000);

		lcd_set_char_pos(0,0);
		lcd_write_string("Testing Encoders");
		
		lcd_set_char_pos(1,0);
		lcd_write_string("Left:");
		lcd_write_uint16(gbl_left_encoder_tics, 0);
		
		lcd_set_char_pos(1,10);
		lcd_write_string("  Rt:");
		lcd_write_uint16(gbl_right_encoder_tics, 0);		

		lcd_set_char_pos(3,3);
		lcd_write_string("Loops:");
		lcd_write_uint16(loops, 0);
	}
}
